from BaseHTTPServer import HTTPServer, BaseHTTPRequestHandler
from SocketServer import ThreadingMixIn
from os import path, mkdir, listdir
from httplib import HTTPResponse
from urlparse import urlparse
from ssl import wrap_socket
from socket import socket
from re import compile
from sys import argv
import re
import threading
import subprocess

class UnsupportedSchemeException(Exception):
    pass

class ProxyHandler(BaseHTTPRequestHandler):

    r = compile(r'http://[^/]+(/?.*)(?i)')

    def _connect_to_host(self):
        # Get hostname and port to connect to
        if self.is_connect:
            host, port = self.path.split(':')
        else:
            u = urlparse(self.path)
            if u.scheme != 'http':
                raise UnsupportedSchemeException('Unknown scheme %s' % repr(u.scheme))
            host = u.hostname
            port = u.port or 80

        # Connect to destination
        self._proxy_sock = socket()
        self._proxy_sock.settimeout(10)
        self._proxy_sock.connect((host, int(port)))

        # Wrap socket if SSL is required
        if self.is_connect:
            self._proxy_sock = wrap_socket(self._proxy_sock)

    def do_CONNECT(self):
        self.is_connect = True
        try:
            # Connect to destination first
            self._connect_to_host()

            # If successful, let's do this!
            self.send_response(200, 'Connection established')
            self.end_headers()
        except Exception, e:
            self.send_error(500, str(e))
            return

        # Reload!
        self.setup()
        self.ssl_host = 'https://%s' % self.path
        self.handle_one_request()


    def do_COMMAND(self):

        # Is this an SSL tunnel?
        path = self.path or '/'

        if not self.is_connect:
            try:
                # Connect to destination
                self._connect_to_host()
            except Exception, e:
                self.send_error(500, str(e))
                return
            # Extract path
            path = self.r.search(path).groups()[0] or '/'

        try:
            # Build request
            req = '%s %s %s\r\n' % (self.command, path, self.request_version)

            # Add headers to the request
            req += '%s\r\n' % self.headers

            # Append message body if present to the request
            if 'Content-Length' in self.headers:
                req += self.rfile.read(int(self.headers['Content-Length']))

            # Send it down the pipe!
            self._proxy_sock.sendall(self.mitm_request(req))

            # Parse response
            h = HTTPResponse(self._proxy_sock)
            h.begin()

            # Get rid of the pesky header
            del h.msg['Transfer-Encoding']

            # Time to relay the message across
            res = '%s %s %s\r\n' % (self.request_version, h.status, h.reason)
            res += '%s\r\n' % h.msg
            res += h.read()

            # Let's close off the remote end
            h.close()
            self._proxy_sock.close()

            # Relay the message
            self.request.sendall(self.mitm_response(res))

        except Exception, e:
            pass

    def mitm_request(self, data):
        for p in self.server._req_plugins:
            data = p(self.server, self).do_request(data)
        return data

    def mitm_response(self, data):
        for p in self.server._res_plugins:
            data = p(self.server, self).do_response(data)
        return data

    def __getattr__(self, item):
        if item.startswith('do_'):
            return self.do_COMMAND


class InterceptorPlugin(object):

    def __init__(self, server, msg):
        self.server = server
        self.message = msg


class RequestInterceptorPlugin(InterceptorPlugin):

    def do_request(self, data):
        return data


class ResponseInterceptorPlugin(InterceptorPlugin):

    def do_response(self, data):
        return data


class InvalidInterceptorPluginException(Exception):
    pass


class MitmProxy(HTTPServer):

    def __init__(self, server_address=('', 8080), RequestHandlerClass=ProxyHandler, bind_and_activate=True):
        HTTPServer.__init__(self, server_address, RequestHandlerClass, bind_and_activate)
        self._res_plugins = []
        self._req_plugins = []

    def register_interceptor(self, interceptor_class):
        if not issubclass(interceptor_class, InterceptorPlugin):
            raise InvalidInterceptorException('Expected type InterceptorPlugin got %s instead' % type(interceptor_class))
        if issubclass(interceptor_class, RequestInterceptorPlugin):
            self._req_plugins.append(interceptor_class)
        if issubclass(interceptor_class, ResponseInterceptorPlugin):
            self._res_plugins.append(interceptor_class)

class BrowserInterceptor(RequestInterceptorPlugin, ResponseInterceptorPlugin):

        def do_request(self, data):
            #print '>> %s' % repr(data[:100])
            #print "..."
            match = re.match("GET \/\/z\/streaming.*_Seg.*\?(pvtoken=[^&]*)&(hdntl=[^&]*)&.*",data)
            if match:
                self.server.pvtoken = match.group(1)
                self.server.hdntl = match.group(2)
                lines = data.splitlines()
                del lines[0]
                for line in lines:
                    match = re.findall("User-Agent: (.*)",line)
                    if match: self.server.userAgent = match[0]
                self.server.stop()
            return data

        def do_response(self, data):
            #print '<< %s' % repr(data[:100])
            return data

class BrowserProxy(ThreadingMixIn, MitmProxy):

    pvtoken   = None
    hdntl     = None
    userAgent = None

    not_stopped = True

    def __init__(self, server_address=('', 8080)):
        MitmProxy.__init__(self, server_address)
        self.register_interceptor(BrowserInterceptor)

    def serve_forever(self):
        """Handle one request at a time until doomsday."""
        while self.not_stopped:
            self.handle_request()

    def stop(self):
        self.not_stopped = False

class ProxyRunner(threading.Thread):

    proxy = None
    def __init__(self, proxy):
        threading.Thread.__init__(self)
        self.proxy = proxy

    def run(self):
        try:
            self.proxy.serve_forever()
        except KeyboardInterrupt:
            self.proxy.server_close()
        return

class BrowserLauncher():

    pvtoken   = None
    hdntl     = None
    userAgent = None

    def __init__(self, browserCmd = "", browserArgs = "", urlEmission = "", port = 8080):
        try:
            proxy = BrowserProxy(('',port))
            runner = ProxyRunner(proxy)
            runner.start()

            args = [browserCmd] + browserArgs.split()
            args.append(urlEmission)
            p = subprocess.Popen(args,shell=False)

            runner.join()
            p.kill()

        except:
            pass

        self.pvtoken = proxy.pvtoken
        self.hdntl = proxy.hdntl
        self.userAgent = proxy.userAgent

if __name__ == '__main__':

    proxy = BrowserProxy()
    try:
        proxy.serve_forever()
    except KeyboardInterrupt:
        proxy.server_close()

